"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const fs = require("fs");
const command_1 = require("@oclif/command");
const nodegit = require("nodegit");
const log_1 = require("../lib/log");
const rules_1 = require("../lib/rules");
class Scan extends command_1.Command {
    async run() {
        var _a;
        const { flags } = this.parse(Scan);
        const ctx = { summary: { findings: [] } };
        const check = async (entry, blobString, ruleType) => {
            var _a;
            const rule = rules_1.rules[ruleType];
            if (!rule || !ctx.currentRepo)
                return;
            const matches = blobString.match(rule.pattern);
            if (matches && matches.length > 0) {
                log_1.default.debug(`[!] found ${matches.length} matches for ${ruleType}`);
                const blame = nodegit.Blame.file(ctx.currentRepo, entry.path());
                for (const matched of matches) {
                    let lineNumber = 0;
                    for (const line of blobString.split("\n")) {
                        lineNumber++;
                        if (line.includes(matched)) {
                            const valuePart = flags.redact ? "" : `found ${matched} `;
                            let blameFinding = undefined;
                            const blameHunk = (await blame).getHunkByLine(lineNumber);
                            const blameCommitId = blameHunk.finalCommitId();
                            try {
                                const blameCommit = await ctx.currentRepo.getCommit(blameCommitId);
                                blameFinding = {
                                    commit: blameCommit.toString(),
                                    author: {
                                        email: blameCommit.committer().email(),
                                    },
                                    date: blameCommit.date(),
                                };
                            }
                            catch (e) {
                                log_1.default.warn(`tried to blame but the related commit ${blameCommitId} has been deleted`);
                            }
                            const finding = {
                                ruleType,
                                detail: `${valuePart}in ${entry.toString()} on line ${lineNumber}`,
                                blame: blameFinding,
                            };
                            (_a = ctx.summary) === null || _a === void 0 ? void 0 : _a.findings.push(finding);
                            log_1.default.warn(`[ ${finding.ruleType} ] ${finding.detail}` +
                                (finding.blame
                                    ? ` (${finding.blame.commit} by ${finding.blame.author.email} at ${finding.blame.date.toISOString()})`
                                    : ""));
                        }
                    }
                }
            }
        };
        const testEntry = async (entry) => {
            var _a;
            log_1.default.debug(`checking ${ctx.currentPath}/${entry.toString()}...`);
            let blob;
            try {
                blob = await entry.getBlob();
            }
            catch (e) {
                log_1.default.debug("no blob data");
                return;
            }
            try {
                const blobString = blob.toString();
                for (const ruleType in rules_1.rules) {
                    const filterEntry = (_a = rules_1.rules[ruleType]) === null || _a === void 0 ? void 0 : _a.filterEntry;
                    if (!filterEntry || filterEntry.test(entry.toString()))
                        await check(entry, blobString, ruleType);
                }
            }
            catch (e) {
                log_1.default.error(e);
            }
        };
        const openRepo = async (dir) => {
            try {
                return await nodegit.Repository.open(path.resolve(dir));
            }
            catch (e) {
                log_1.default.error("failed to open repository");
                throw e;
            }
        };
        const fetchAll = async (repo) => await repo.fetchAll({
            callbacks: {
                certificateCheck: () => 0,
                credentials: (_url, user) => {
                    log_1.default.debug(`fetching from ${_url}`);
                    return nodegit.Cred.sshKeyFromAgent(user);
                },
            },
        });
        const latestCommit = async (repo) => {
            try {
                return await repo.getHeadCommit();
            }
            catch (e) {
                log_1.default.error("failed to get latest HEAD commit");
                throw e;
            }
        };
        const checkTreeEntries = async (entries) => {
            for (const entry of entries) {
                if (entry.isTree())
                    await checkTreeEntries((await entry.getTree()).entries());
                await testEntry(entry);
            }
        };
        const scanRepoDir = async (dir) => {
            log_1.default.info(`scanning ${dir}`);
            try {
                ctx.currentRepo = await openRepo(dir);
                ctx.currentPath = dir;
                if (flags.pull) {
                    log_1.default.info("fetching from remotes...");
                    await fetchAll(ctx.currentRepo);
                }
                const commit = await latestCommit(ctx.currentRepo);
                const branchName = (await ctx.currentRepo.getCurrentBranch()).shorthand();
                log_1.default.info(`analysing files at commit ${commit.toString()} (${branchName})`);
                const tree = await commit.getTree();
                await checkTreeEntries(tree.entries());
            }
            catch (e) {
                log_1.default.warn("directory not scanned - check debug log for details");
                log_1.default.debug(e);
            }
        };
        try {
            const resolvedDir = path.resolve(flags.dir || ".");
            // is this a git repo?
            const gitDir = path.resolve(resolvedDir, ".git");
            log_1.default.info(`SCAN STARTED: ${resolvedDir}`);
            log_1.default.debug(`checking for ${gitDir}`);
            if (!fs.existsSync(gitDir)) {
                // not a git repo, iterate directories
                log_1.default.debug(`${resolvedDir} is not a git repo, digging...`);
                const files = await fs.promises.readdir(resolvedDir);
                for (const file of files) {
                    const fullPath = path.resolve(resolvedDir, file);
                    const stat = await fs.promises.stat(fullPath);
                    if (stat.isDirectory())
                        await scanRepoDir(fullPath);
                }
            }
            else
                await scanRepoDir(resolvedDir);
        }
        catch (e) {
            log_1.default.error(e, "SCAN FAILED");
            this.exit(1);
        }
        log_1.default.info(`SCAN COMPLETE (found ${(_a = ctx.summary) === null || _a === void 0 ? void 0 : _a.findings.length} issues)`);
    }
}
exports.default = Scan;
Scan.description = "scans local git repositories for committed secrets";
Scan.examples = ["$ git-secret-scanner scan ~/code/github.com/my-org -r"];
Scan.flags = {
    help: command_1.flags.help({ char: "h" }),
    dir: command_1.flags.string({
        char: "d",
        description: "directory to scan (current direction if omitted)",
    }),
    pull: command_1.flags.boolean({ char: "p", description: "pull from repositories" }),
    redact: command_1.flags.boolean({
        char: "r",
        description: "redact all matched secret strings",
    }),
};
